from collections import namedtuple
from typing import Optional


def match(data: dict, pattern: dict) -> Optional[namedtuple]:  # noqa PYI024
    """
    Check if a dictionary (`data`) matches a specified pattern (`pattern`).

    This function verifies if the structure and types of `data` conform to those
    defined in `pattern`. If they match, a named tuple containing the matching values
    is returned. If not, the function returns `None`.

    :param data: The dictionary to be checked against the pattern.
    :param pattern: The dictionary defining the structure and expected types for matching.
    :return: A named tuple with the matching values if `data` matches `pattern`. `None` if there is no match.
    :raise TypeError: If a type is used as a key in the `pattern` dictionary.
    """
    if not (isinstance(data, dict) or isinstance(pattern, dict)):
        return None

    if len(data) != len(pattern):
        return None

    result_data: dict = {}
    for pattern_key, pattern_value in pattern.items():
        if isinstance(pattern_key, type):
            message: str = f"Can't use types as pattern keys: {pattern}. Key: {pattern_key}."
            raise TypeError(message)

        if pattern_key not in data:
            return None

        value = data[pattern_key]
        if isinstance(pattern_value, type) and isinstance(value, pattern_value):
            result_data[pattern_key] = value
        elif isinstance(pattern_value, dict) and (mr := match(value, pattern_value)):
            result_data[pattern_key] = mr
        elif pattern_value == value:
            result_data[pattern_key] = value
        else:
            return None

    return namedtuple(typename="MatchResult", field_names=result_data.keys())(*result_data.values())  # noqa PYI024



from pathlib import Path
from typing import Optional

from kataloger.data.artifact.artifact import Artifact
from kataloger.data.artifact.library import Library
from kataloger.data.artifact.plugin import Plugin
from kataloger.data.artifact_update import ArtifactUpdate
from kataloger.data.metadata_repository_info import MetadataRepositoryInfo
from kataloger.data.repository import Repository
from kataloger.exceptions.kataloger_configuration_exception import KatalogerConfigurationError
from kataloger.helpers.log_helpers import log_warning
from kataloger.helpers.toml_parse_helpers import load_catalog
from kataloger.helpers.update_helpers import get_all_artifact_metadata
from kataloger.update_resolver.base.update_resolution import UpdateResolution
from kataloger.update_resolver.base.update_resolver import UpdateResolver


class CatalogUpdater:
    def __init__(
        self,
        library_repositories: list[Repository],
        plugin_repositories: list[Repository],
        update_resolvers: list[UpdateResolver],
        *,
        verbose: bool = False,
    ):
        if not (library_repositories or plugin_repositories):
            message = "No repositories provided!"
            raise KatalogerConfigurationError(message)

        if not update_resolvers:
            message = "No update resolvers provided!"
            raise KatalogerConfigurationError(message)

        self.library_repositories = library_repositories
        self.plugin_repositories = plugin_repositories
        self.update_resolvers = update_resolvers
        self.verbose = verbose

    async def get_catalog_updates(self, catalog_path: Path) -> list[ArtifactUpdate]:
        libraries, plugins = load_catalog(catalog_path, verbose=self.verbose)
        if not (libraries or plugins):
            if self.verbose:
                log_warning(f'Catalog "{catalog_path.name}" is empty.')
            return []

        library_updates, plugin_updates = await self.get_updates(libraries, plugins)
        return library_updates + plugin_updates

    async def get_artifact_updates(self, artifacts: list[Artifact]) -> list[ArtifactUpdate]:
        libraries = [artifact for artifact in artifacts if isinstance(artifact, Library)]
        plugins = [artifact for artifact in artifacts if isinstance(artifact, Plugin)]
        library_updates, plugin_updates = await self.get_updates(libraries, plugins)
        return library_updates + plugin_updates

    async def get_updates(
        self,
        libraries: list[Library],
        plugins: list[Plugin],
    ) -> tuple[list[ArtifactUpdate], list[ArtifactUpdate]]:
        library_updates = await self.get_library_updates(libraries)
        plugin_updates = await self.get_plugin_updates(plugins)
        return library_updates, plugin_updates

    async def get_library_updates(self, libraries: list[Library]) -> list[ArtifactUpdate]:
        library_updates: list[ArtifactUpdate] = []
        if not self.library_repositories:
            if self.verbose:
                log_warning("No repositories for libraries provided.")
            return library_updates

        library_update_info = await get_all_artifact_metadata(
            artifacts=libraries,
            repositories=self.library_repositories,
            verbose=self.verbose,
        )
        for artifact, repositories_metadata in library_update_info.items():
            library_updates.append(self.try_find_update(artifact, repositories_metadata))
        return list(filter(lambda item: item is not None, library_updates))

    async def get_plugin_updates(self, plugins: list[Plugin]) -> list[ArtifactUpdate]:
        plugin_updates: list[ArtifactUpdate] = []
        if not self.plugin_repositories:
            if self.verbose:
                log_warning("No repositories for plugins provided.")
            return plugin_updates

        plugin_update_info = await get_all_artifact_metadata(
            artifacts=plugins,
            repositories=self.plugin_repositories,
            verbose=self.verbose,
        )
        for artifact, repositories_metadata in plugin_update_info.items():
            plugin_updates.append(self.try_find_update(artifact, repositories_metadata))
        return list(filter(lambda item: item is not None, plugin_updates))

    def try_find_update(
        self,
        artifact: Artifact,
        repositories_metadata: list[MetadataRepositoryInfo],
    ) -> Optional[ArtifactUpdate]:
        for resolver in self.update_resolvers:
            (resolution, optional_update) = resolver.resolve(artifact, repositories_metadata)
            if resolution == UpdateResolution.CANT_RESOLVE:
                continue
            if resolution == UpdateResolution.UPDATE_FOUND:
                return optional_update
            if resolution == UpdateResolution.NO_UPDATES:
                return None

            message: str = f'Unexpected update resolution: "{resolution}".'
            raise ValueError(message)

        return None



from typing import Optional

from kataloger.data.artifact.artifact import Artifact
from kataloger.data.artifact_update import ArtifactUpdate
from kataloger.data.metadata_repository_info import MetadataRepositoryInfo
from kataloger.update_resolver.base.update_resolution import UpdateResolution
from kataloger.update_resolver.base.update_resolver import UpdateResolver
from kataloger.update_resolver.universal.version import Version
from kataloger.update_resolver.universal.version_factory import VersionFactory


class UniversalUpdateResolver(UpdateResolver):

    def __init__(
        self,
        version_factories: list[VersionFactory[Version]],
        *,
        suggest_unstable_updates: bool,
    ):
        self.version_factories = version_factories
        self.suggest_unstable_updates = suggest_unstable_updates

    def resolve(
        self,
        artifact: Artifact,
        repositories_metadata: list[MetadataRepositoryInfo],
    ) -> tuple[UpdateResolution, Optional[ArtifactUpdate]]:
        recently_updated_repository = self.__most_recently_updated_repository(repositories_metadata)
        current_version_repository = self.__repository_with_current_version(artifact.version, repositories_metadata)
        repositories_to_check: list[MetadataRepositoryInfo] = []
        if current_version_repository and recently_updated_repository != current_version_repository:
            repositories_to_check.append(current_version_repository)
        repositories_to_check.append(recently_updated_repository)

        for repository in repositories_to_check:
            for factory in self.version_factories:
                (resolution, optional_update) = self.__resolve_update_in_repository(artifact, factory, repository)
                if resolution == UpdateResolution.CANT_RESOLVE:
                    continue
                if resolution == UpdateResolution.NO_UPDATES or resolution == UpdateResolution.UPDATE_FOUND:
                    return resolution, optional_update

                message: str = f'Unexpected update resolution: "{resolution}".'
                raise ValueError(message)

        return UpdateResolution.CANT_RESOLVE, None

    def __resolve_update_in_repository(
        self,
        artifact: Artifact,
        version_factory: VersionFactory,
        repository_metadata: MetadataRepositoryInfo,
    ) -> tuple[UpdateResolution, Optional[ArtifactUpdate]]:
        current_version = artifact.version
        if not version_factory.can_create(current_version):
            return UpdateResolution.CANT_RESOLVE, None

        artifact_version = version_factory.create(current_version)
        suggest_unstable = self.suggest_unstable_updates or artifact_version.is_pre_release()
        for version in reversed(repository_metadata.metadata.versions):
            if not version_factory.can_create(version):
                return UpdateResolution.CANT_RESOLVE, None

            update_version = version_factory.create(version)
            if update_version.is_pre_release() and not suggest_unstable:
                continue

            if artifact_version == update_version:
                return UpdateResolution.NO_UPDATES, None
            if artifact_version < update_version:
                update = ArtifactUpdate(
                    name=artifact.name,
                    update_repository_name=repository_metadata.repository.name,
                    current_version=current_version,
                    available_version=version,
                )
                return UpdateResolution.UPDATE_FOUND, update

        return UpdateResolution.CANT_RESOLVE, None

    @staticmethod
    def __most_recently_updated_repository(
        repositories_metadata: list[MetadataRepositoryInfo],
    ) -> MetadataRepositoryInfo:
        return max(repositories_metadata, key=lambda rm: rm.metadata.last_updated)

    @staticmethod
    def __repository_with_current_version(
        current_version: str,
        repositories_metadata: list[MetadataRepositoryInfo],
    ) -> Optional[MetadataRepositoryInfo]:
        for repository_metadata in repositories_metadata:
            if current_version in repository_metadata.metadata.versions:
                return repository_metadata
        return None



from dataclasses import dataclass
from typing import Optional

from kataloger.data.catalog import Catalog
from kataloger.data.repository import Repository


@dataclass(frozen=True)
class ConfigurationData:
    catalogs: Optional[list[Catalog]]
    library_repositories: Optional[list[Repository]]
    plugin_repositories: Optional[list[Repository]]
    verbose: Optional[bool]
    suggest_unstable_updates: Optional[bool]
    fail_on_updates: Optional[bool]



from dataclasses import dataclass


@dataclass(frozen=True)
class ArtifactMetadata:
    latest_version: str
    release_version: str
    versions: list[str]
    last_updated: int

    def __repr__(self):
        return f"{len(self.versions)}/{self.latest_version}"



from dataclasses import dataclass

from kataloger.data.catalog import Catalog
from kataloger.data.repository import Repository


@dataclass(frozen=True)
class KatalogerConfiguration:
    catalogs: list[Catalog]
    library_repositories: list[Repository]
    plugin_repositories: list[Repository]
    verbose: bool
    suggest_unstable_updates: bool
    fail_on_updates: bool



from kataloger.data.artifact_update import ArtifactUpdate


def print_catalog_updates(
    updates: list[ArtifactUpdate],
    catalog_name: str,
    catalog_count: int,
    *,
    verbose: bool,
) -> None:
    if catalog_count > 1:
        if updates:
            print(f'Updates for "{catalog_name}" catalog:')
        else:
            print(f'Catalog "{catalog_name}" is up to date!')

    for update in updates:
        version_part = f"{update.current_version} -> {update.available_version}"
        if verbose:
            print(f"[{update.update_repository_name}] {update.name} {version_part}")
        else:
            print(f"{update.name} {version_part}")

    if catalog_count > 1:
        print()



from dataclasses import dataclass


@dataclass
class ArtifactUpdate:
    name: str
    update_repository_name: str
    current_version: str
    available_version: str

    def __repr__(self):
        return f"{self.name} {self.current_version} -> {self.available_version}"



from argparse import ArgumentParser
from pathlib import Path
from typing import Optional

from kataloger import __version__ as package_version
from kataloger.data.catalog import Catalog
from kataloger.data.configuration_data import ConfigurationData
from kataloger.data.kataloger_arguments import KatalogerArguments
from kataloger.helpers.path_helpers import str_to_path


def parse_arguments(*args: str) -> KatalogerArguments:
    parser = ArgumentParser(
        prog="kataloger",
        description="A Python command-line utility to discover updates for your gradle version catalogs.",
        allow_abbrev=False,
        epilog="Visit project repository to get more information.",
    )
    parser.add_argument(
        "-p",
        "--path",
        action="append",
        dest="paths",
        help="Path(s) to gradle version catalog. If catalog path not provided script looking for "
             "version catalogs in current directory.",
    )
    parser.add_argument(
        "-c",
        "--configuration",
        type=str,
        dest="configuration_path",
        metavar="path",
        help="Path to .toml file with configuration. If path not provided script looking for "
             "default.configuration.toml file in current directory.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=None,
        dest="verbose",
        help="Enables detailed output.",
    )
    parser.add_argument(
        "-u",
        "--suggest-unstable",
        action="store_true",
        default=None,
        dest="suggest_unstable_updates",
        help="Allow %(prog)s suggest update from stable version to unstable.",
    )
    parser.add_argument(
        "-f",
        "--fail-on-updates",
        action="store_true",
        default=None,
        dest="fail_on_updates",
        help="Exit with non-zero code when at least one update found.",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {package_version}",
    )
    arguments = parser.parse_args(args)

    return KatalogerArguments(
        configuration_path=_get_configuration_path(arguments.configuration_path),
        configuration_data=ConfigurationData(
            catalogs=_get_catalogs(arguments.paths),
            library_repositories=None,
            plugin_repositories=None,
            verbose=arguments.verbose,
            suggest_unstable_updates=arguments.suggest_unstable_updates,
            fail_on_updates=arguments.fail_on_updates,
        ),
    )


def _get_catalogs(path_strings: list[str]) -> Optional[list[Catalog]]:
    if path_strings:
        return [Catalog.from_path(str_to_path(path_string=path_str, root_path=Path.cwd())) for path_str in path_strings]

    return None


def _get_configuration_path(path_string: Optional[str]) -> Optional[Path]:
    if path_string:
        return str_to_path(path_string=path_string, root_path=Path.cwd())

    return None



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from kataloger.update_resolver.universal.version import Version

T = TypeVar("T", bound=Version)


class VersionFactory(ABC, Generic[T]):

    @abstractmethod
    def create(self, version: str) -> T:
        raise NotImplementedError

    @abstractmethod
    def can_create(self, version: str) -> bool:
        raise NotImplementedError



from dataclasses import dataclass
from typing import Optional

from yarl import URL


@dataclass(frozen=True)
class Repository:
    name: str
    address: URL
    user: Optional[str] = None
    password: Optional[str] = None

    def __repr__(self):
        return self.name

    def requires_authorization(self) -> bool:
        return self.user is not None and self.password is not None



from pyexpat import ExpatError
from typing import Optional

import xmltodict

from kataloger.data.artifact_metadata import ArtifactMetadata


def try_parse_maven_group_metadata(response: str) -> Optional[ArtifactMetadata]:
    try:
        metadata = xmltodict.parse(response.strip())
    except ExpatError:
        return None

    try:
        version_info = metadata["metadata"]["versioning"]
        versions = version_info["versions"]["version"]
        if not isinstance(versions, list):
            versions = [versions]

        return ArtifactMetadata(
            latest_version=version_info.get("latest", versions[-1]),
            release_version=version_info.get("release", versions[-1]),
            versions=versions,
            last_updated=int(version_info.get("lastUpdated", 0)),
        )
    except (KeyError, TypeError):
        return None



from importlib.resources import as_file, files
from pathlib import Path
from typing import Optional

from kataloger import package_name
from kataloger.exceptions.kataloger_configuration_exception import KatalogerConfigurationError


def str_to_path(path_string: str, root_path: Optional[Path] = None) -> Path:
    """
    Converts a string representing a file path into a Path object. It expands user tilde (~) and resolves relative
    paths against a provided base path. It also verifies the existence of the file at the specified path.

    :param path_string: A string representing the file path to be converted.
    :param root_path: An optional base path against which to resolve relative paths. Defaults to None.
    :return: A Path object representing the absolute file path.
    :raise KatalogerConfigurationException: If the resolved path does not point to an existing file.
    """
    path = Path(path_string).expanduser()
    if not path.is_absolute() and root_path:
        path = (root_path / path).resolve()

    if not file_exists(path):
        raise KatalogerConfigurationError(message=f'Incorrect path: "{path_string}".')

    return path


def file_exists(path: Path) -> bool:
    """
    Checks whether a given path exists and is a file.

    :param path: A Path object representing the file path to be checked.
    :returns: True if the path exists and is a file, False otherwise.
    """
    return path.exists() and path.is_file()


def get_package_file(filename: str) -> Path:
    """
    Returns the absolute path to a file inside the installed package.

    :param filename: The file name within the package.
    :returns: The absolute path to the file.
    """
    with as_file(files(package_name).joinpath(filename)) as path:
        return path



import re
from functools import total_ordering
from itertools import zip_longest

from kataloger.update_resolver.universal.version import Version


@total_ordering
class UniversalVersion(Version):
    __regex = re.compile(r"^([0-9]{,10}(?:\.[0-9]{1,10}){,20})(?:[-.]([a-zA-Z.-]{1,100})([0-9]{1,10})?)?$")

    __pre_release_names = ("dev", "alpha", "beta", "rc")

    def __init__(self, version: str):
        super().__init__(version)
        match = self.__regex.match(version)
        self.numeric_part: str = match.group(1)
        self.pre_release_name: str = match.group(2)

        pre_release_number = match.group(3)
        if pre_release_number is None:
            pre_release_number = 0
        self.pre_release_number: int = int(pre_release_number)

    def is_pre_release(self) -> bool:
        return self.pre_release_name is not None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, UniversalVersion):
            return False

        return self.raw == other.raw

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, UniversalVersion):
            return False

        if self == other:
            return False

        def to_digits_list(version_part: str) -> list[int]:
            return list(map(int, version_part.split(".")))

        # Compare numeric part of version. Missing digits assumed as 0.
        self_digits = to_digits_list(self.numeric_part)
        other_digits = to_digits_list(other.numeric_part)
        for self_digit, other_digit in zip_longest(self_digits, other_digits, fillvalue=0):
            if self_digit < other_digit:
                return True
            if self_digit > other_digit:
                return False

        # Numeric part is equal, lets compare pre-release part.
        if self.is_pre_release() and other.is_pre_release():
            if self.pre_release_name == other.pre_release_name:
                # Pre-release prefixes are equals, compare pre-release numbers
                if self.pre_release_number < other.pre_release_number:
                    return True
            else:
                # Pre-release prefixes are different, compare pre-release prefixes
                self_pr_index = self._pre_release_index()
                other_pr_index = other._pre_release_index()
                if self_pr_index < other_pr_index:
                    return True
                if self_pr_index > other_pr_index:
                    return False
            return False

        return self.is_pre_release() and not other.is_pre_release()

    def _pre_release_index(self) -> int:
        lowercase_pre_release_name = self.pre_release_name.lower()
        if lowercase_pre_release_name in self.__pre_release_names:
            return self.__pre_release_names.index(lowercase_pre_release_name)
        return -1

    @classmethod
    def can_handle(cls, version: str) -> bool:
        return cls.__regex.match(version) is not None



from pathlib import Path
from typing import Optional, Union

from yarl import URL

from kataloger.data.artifact.library import Library
from kataloger.data.artifact.plugin import Plugin
from kataloger.data.catalog import Catalog
from kataloger.data.configuration_data import ConfigurationData
from kataloger.data.repository import Repository
from kataloger.exceptions.kataloger_parse_exception import KatalogerParseError
from kataloger.helpers.backport_helpers import load_toml
from kataloger.helpers.log_helpers import log_warning
from kataloger.helpers.path_helpers import str_to_path
from kataloger.helpers.structural_matching_helpers import match


def load_configuration(configuration_path: Path) -> ConfigurationData:
    catalogs: Optional[list[Catalog]] = None
    library_repositories: Optional[list[Repository]] = None
    plugin_repositories: Optional[list[Repository]] = None

    configuration_data = load_toml(path=configuration_path)
    if "catalogs" in configuration_data:
        catalogs = parse_catalogs(configuration_data["catalogs"], configuration_root_dir=configuration_path.parent)
    if "libraries" in configuration_data:
        library_repositories = parse_repositories(configuration_data["libraries"])
    if "plugins" in configuration_data:
        plugin_repositories = parse_repositories(configuration_data["plugins"])

    return ConfigurationData(
        catalogs=catalogs,
        library_repositories=library_repositories,
        plugin_repositories=plugin_repositories,
        verbose=__extract_optional_boolean(configuration_data, key="verbose"),
        suggest_unstable_updates=__extract_optional_boolean(configuration_data, key="suggest_unstable_updates"),
        fail_on_updates=__extract_optional_boolean(configuration_data, key="fail_on_updates"),
    )


def load_catalog(catalog_path: Path, *, verbose: bool) -> tuple[list[Library], list[Plugin]]:
    catalog = load_toml(catalog_path)
    versions: dict[str, str] = catalog.pop("versions", {})
    libraries = parse_libraries(catalog, versions, verbose=verbose)
    plugins = parse_plugins(catalog, versions, verbose=verbose)

    return libraries, plugins


def parse_repositories(data: dict) -> Optional[list[Repository]]:
    if not data:
        return None

    if not isinstance(data, dict):
        raise KatalogerParseError(message="Unexpected repository data.")

    repositories = []
    for name, repository_data in data.items():
        if not isinstance(name, str):
            raise KatalogerParseError(message=f'Unexpected repository name: "{name}".')

        repository: Repository
        if isinstance(repository_data, str):
            repository = Repository(name=name, address=URL(repository_data))
        elif mr := match(repository_data, pattern={"address": str, "user": str, "password": str}):
            repository = Repository(
                name=name,
                address=URL(mr.address),
                user=mr.user,
                password=mr.password,
            )
        else:
            raise KatalogerParseError(message="Unexpected repository data.")
        repositories.append(repository)

    return repositories


def parse_catalogs(data: Union[list, dict], configuration_root_dir: Optional[Path]) -> Optional[list[Catalog]]:
    if not data:
        return None

    catalogs = []
    if isinstance(data, list):
        for path in data:
            if not isinstance(path, str):
                raise KatalogerParseError(message=f'Unexpected catalog path: "{path}".')
            if not path.strip():
                raise KatalogerParseError(message="Catalog path can't be empty!")

            catalogs.append(Catalog.from_path(path=str_to_path(path, root_path=configuration_root_dir)))
        return catalogs

    if isinstance(data, dict):
        for name, path in data.items():
            if not isinstance(name, str):
                raise KatalogerParseError(message=f'Unexpected catalog name: "{name}".')
            if not isinstance(path, str):
                raise KatalogerParseError(message=f'Unexpected catalog path: "{path}".')
            if not name.strip():
                raise KatalogerParseError(message="Catalog name can't be empty!")
            if not path.strip():
                raise KatalogerParseError(message=f'Catalog "{name}" has empty path!')

            catalogs.append(Catalog(name=name, path=str_to_path(path, root_path=configuration_root_dir)))
        return catalogs

    raise KatalogerParseError(message=f'Unexpected catalogs data format: "{data}".')


def parse_libraries(
    catalog: dict[str,Union[dict, str]],
    versions: dict,
    *,
    verbose: bool,
) -> list[Library]:
    libraries = []
    if "libraries" not in catalog:
        return libraries

    for name, library_data in catalog["libraries"].items():
        if not isinstance(name, str):
            raise KatalogerParseError(message=f'Unexpected library name: "{name}".')

        library: Library
        if isinstance(library_data, str):
            (module, version) = __parse_declaration(library_data)
            library = Library(name=name, coordinates=module, version=version)
        elif mr := match(library_data, pattern={"group": str, "name": str, "version": {"ref": str}}):
            library = Library(
                name=name,
                coordinates=f"{mr.group}:{mr.name}",
                version=__get_version_by_reference(versions, mr.version.ref, name),
            )
        elif mr := match(library_data, pattern={"group": str, "name": str, "version": str}):
            library = Library(
                name=name,
                coordinates=f"{mr.group}:{mr.name}",
                version=mr.version,
            )
        elif mr := match(library_data, pattern={"group": str, "name": str}):
            if verbose:
                log_warning(f'Library "{mr.module}" has no version in catalog.')
            continue
        elif mr := match(library_data, pattern={"module": str, "version": {"ref": str}}):
            library = Library(
                name=name,
                coordinates=mr.module,
                version=__get_version_by_reference(versions, mr.version.ref, name),
            )
        elif mr := match(library_data, pattern={"module": str, "version": str}):
            library = Library(
                name=name,
                coordinates=mr.module,
                version=mr.version,
            )
        elif mr := match(library_data, pattern={"module": str}):
            if verbose:
                log_warning(f'Library "{mr.module}" has no version in catalog.')
            continue
        else:
            message = f"Unknown library notation: {library_data}"
            raise KatalogerParseError(message)

        libraries.append(library)

    return libraries


def parse_plugins(
    catalog: dict[str, Union[dict, str]],
    versions: dict,
    *,
    verbose: bool,
) -> list[Plugin]:
    plugins = []
    if "plugins" not in catalog:
        return plugins

    for name, plugin_data in catalog["plugins"].items():
        if not isinstance(name, str):
            raise KatalogerParseError(message=f'Unexpected plugin name: "{name}".')

        if isinstance(plugin_data, str):
            (plugin_id, version) = __parse_declaration(plugin_data)
            plugin = Plugin(name=name, coordinates=plugin_id, version=version)
        elif mr := match(plugin_data, pattern={"id": str, "version": str}):
            plugin = Plugin(
                name=name,
                coordinates=mr.id,
                version=mr.version,
            )
        elif mr := match(plugin_data, pattern={"id": str, "version": {"ref": str}}):
            plugin = Plugin(
                name=name,
                coordinates=mr.id,
                version=__get_version_by_reference(versions, mr.version.ref, name),
            )
        elif mr := match(plugin_data, pattern={"id": str}):
            if verbose:
                log_warning(f'Plugin "{mr.id}" has no version in catalog.')
            continue
        else:
            message = f"Unknown plugin notation: {plugin_data}"
            raise KatalogerParseError(message)

        plugins.append(plugin)

    return plugins


def __parse_declaration(declaration: str) -> tuple[str, str]:
    components = declaration.rsplit(":", 1)
    if len(components) != 2 or not (components[0].strip() and components[1].strip()):
        message = f'Unknown declaration format: "{declaration}".'
        raise KatalogerParseError(message)
    return components[0], components[1]


def __get_version_by_reference(
    versions: dict[str, str],
    version_ref: str,
    artifact_name: str,
) -> str:
    if not (version := versions.get(version_ref)):
        message = f'Version for "{artifact_name}" not specified by reference "{version_ref}".'
        raise KatalogerParseError(message)

    return version


def __extract_optional_boolean(data: dict, key: str) -> Optional[bool]:
    value = data.get(key)
    if value is None or isinstance(value, bool):
        return value

    message = f'Configuration field "{key}" has incorrect value "{value}", while expected boolean type.'
    raise KatalogerParseError(message)



import asyncio
import sys

from kataloger.cli import cli
from kataloger.exceptions.kataloger_exception import KatalogerError


def main() -> int:
    try:
        return asyncio.run(cli.run())
    except KatalogerError as error:
        print(error.message, file=sys.stderr)
        return 1
    except KeyboardInterrupt:
        print("Update search terminated.")
        return 128 + 2  # http://www.tldp.org/LDP/abs/html/exitcodes.html


if __name__ == "__main__":
    sys.exit(main())



import sys
from pathlib import Path
from typing import Union

from kataloger.exceptions.kataloger_parse_exception import KatalogerParseError


def load_toml(path: Path) -> dict[str, Union[str, dict]]:
    if sys.version_info < (3, 11):
        import tomli as tomllib
        from tomli import TOMLDecodeError
    else:
        import tomllib
        from tomllib import TOMLDecodeError

    with Path.open(path, mode="rb") as file:
        try:
            return tomllib.load(file)
        except TOMLDecodeError as parse_error:
            message = f"Can't parse TOML in \"{path.name}\"."
            raise KatalogerParseError(message) from parse_error



from kataloger.catalog_updater import CatalogUpdater
from kataloger.cli.configuration_provider import get_configuration
from kataloger.cli.update_print_helper import print_catalog_updates
from kataloger.update_resolver.universal.universal_update_resolver import UniversalUpdateResolver
from kataloger.update_resolver.universal.universal_version_factory import UniversalVersionFactory


async def run() -> int:
    configuration = get_configuration()

    update_resolver = UniversalUpdateResolver(
        version_factories=[UniversalVersionFactory()],
        suggest_unstable_updates=configuration.suggest_unstable_updates,
    )

    catalog_updater = CatalogUpdater(
        library_repositories=configuration.library_repositories,
        plugin_repositories=configuration.plugin_repositories,
        update_resolvers=[update_resolver],
        verbose=configuration.verbose,
    )

    has_updates = False
    for catalog in configuration.catalogs:
        updates = await catalog_updater.get_catalog_updates(catalog.path)
        if not has_updates and updates:
            has_updates = True

        print_catalog_updates(
            updates=updates,
            catalog_name=catalog.name,
            catalog_count=len(configuration.catalogs),
            verbose=configuration.verbose,
        )

    if configuration.fail_on_updates and has_updates:
        return 1
    return 0
